#
# Copyright (C) 2013-2013, Intercom Srl, Daniele Orlandi
#
# Author:: Daniele Orlandi <daniele@orlandi.com>
#          Lele Forzani <lele@windmill.it>
#          Alfredo Cerutti <acerutti@intercom.it>
#
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.
#

require 'kyotocabinet'
require 'ygg/provisioner/tools'
require 'ygg/provisioner/file_switcher'

module Ygg
module Provisioner

class Model < OpenStruct

  class Error < StandardError ; end
  class DataError < Error ; end
  class ProcessError < Error ; end
  class DatabaseError < Error ; end

  extend Tools
  include Tools
  include FileSwitcher

  class_attribute :agentcfg
  class_attribute :log
  class_attribute :config
  class_attribute :worker

  class_attribute :state_persistant_keys
  self.state_persistant_keys = [ ]

  class_attribute :db_path

  attr_accessor :plog
  attr_accessor :config_persistant_keys

  def initialize(config)

    super(*[])

    self.cfg = agentcfg

    if config[:cfg]
      # We are being de-serialized
      update_config(config[:cfg])
      update_state(config[:state])
    else
      update_config(config)
    end
  end

  def update_config(config)
    self.config_persistant_keys = config.keys.map(&:to_sym)
    self.config_persistant_keys -= self.state_persistant_keys
    config.each { |k,v| send("#{k}=", v) if self.config_persistant_keys.include?(k.to_sym) }
  end

  def update_state(state)
    state.each { |k,v| send("#{k}=", v) if self.state_persistant_keys.include?(k.to_sym) }
  end

  def save!
    self.class.save(self)
  end

  def serialize
   {
    :cfg => self.marshal_dump.select { |k,v| config_persistant_keys.include?(k.to_sym) },
    :state => self.marshal_dump.select { |k,v| state_persistant_keys.include?(k.to_sym) },
    :_type => self.class.name,
   }
  end

  def logging_on(plog)
    @plog = plog
    yield
  end

  class NotFound < StandardError ; end

  @@open_store = false

  def self.with_open_store(&block)
    if @@open_store
      yield @@open_store
    else
      db = KyotoCabinet::DB::new
      if db.open(db_path, KyotoCabinet::DB::OWRITER | KyotoCabinet::DB::OCREATE)
        begin
          @@open_store = db
          yield db
        ensure
          unless db.close
            raise DatabaseError.new("Cannot close database #{db_path}: #{db.error}")
          end
          @@open_store = nil
        end
      else
        raise DatabaseError.new("Cannot open database #{db_path}: #{db.error}")
      end
    end
  end

  def self.save(model)
    with_open_store do |store|
      store[model.id] = Marshal.dump(model.serialize)
    end

    model
  end

  def self.new_from_serialized(ser_obj)
    klass = self
    klass = ser_obj[:_type].constantize if ser_obj[:_type]
    obj = klass.new(ser_obj)
    obj
  end

  def self.each
    with_open_store do |store|
      store.each do |obj_id,obj|
        begin
          obj = new_from_serialized(Marshal.load(obj))

          yield obj

#        rescue
#          # Ignore corrupted obj data
#          # plog.warn "Site #{obj_id} configuration is corrupted"
        end
      end
    end
  end

  def self.find(id)
    obj = nil
    with_open_store do |store|
      obj = store[id]
      raise NotFound if !obj
      obj = new_from_serialized(Marshal.load(obj))
    end

    obj
  end

  def self.delete(id)
    obj = nil
    with_open_store do |store|
      obj = store[id]
      raise NotFound if !obj
      store.delete id
    end
  end

  def self.count
    res = nil

    with_open_store do |store|
      res = store.size
    end

    res
  end

  def self.fix_database(opts = {})
    corrupted_objs = {}
    good_count = 0

    with_open_store do |store|
      store.each do |obj_id,obj_marshalled|
        begin
          obj_unmarshalled = nil

          begin
            obj_unmarshalled = Marshal.load(obj_marshalled)
          rescue Exception => e
            corrupted_objs[obj_id] = { :reason => 'Unmarshal failed', :data => Base64.encode64(obj_marshalled) }
            raise
          end

          obj = nil
          begin
            obj = new_from_serialized(obj_unmarshalled)
          rescue Exception => e
            corrupted_objs[obj_id] = {
              :reason => "Loading failed: #{e.to_s}",
              :data => obj_unmarshalled
            }
            raise
          end

          begin
            obj.save!
            good_count += 1
          rescue Exception => e
            corrupted_objs[obj_id] = {
              :reason => "Saving failed: #{e.to_s}",
              :data => obj_unmarshalled
            }
            raise
          end
        rescue
          if opts[:delete_corrupted]
            store.delete obj_id
            corrupted_objs[obj_id][:deleted] = true
          end
        end
      end
    end

    {
     :good_count => good_count,
     :corrupted_objs => corrupted_objs
    }
  end
end

end
end
