#
# Copyright (C) 2013-2013, Intercom Srl, Daniele Orlandi
#
# Author:: Daniele Orlandi <daniele@orlandi.com>
#          Lele Forzani <lele@windmill.it>
#          Alfredo Cerutti <acerutti@intercom.it>
#
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.
#

require 'fiber'

require 'ygg/provisioner/tools'
require 'ygg/provisioner/file_switcher'

module Ygg
module Provisioner

class BufferedLog
  attr_accessor :text
  attr_accessor :other_log

  def initialize(other_log)
    @text = ''
    @other_log = other_log
  end

  def log(level, msg)
    @text << msg + "\n"
    @other_log.send level, msg
  end

  def debug(msg) ; log :debug, msg ; end
  def info(msg) ; log :info, msg ; end
  def warn(msg) ; log :warn, msg ; end
  def error(msg) ; log :error, msg ; end
  def crit(msg) ; log :crit, msg ; end
end

class Task
  class PermanentFailure < StandardError ; end
  class TemporaryFailure < StandardError ; end
  class PublishError < StandardError ; end
  class InvalidRequestData < PermanentFailure ; end
  class OperationNotDefined < StandardError ; end

  extend Tools
  include Tools
  include FileSwitcher

  class Operation
    attr_reader :help_caption
    attr_reader :help_text
    attr_reader :supports_percent

    def initialize
      @help_caption = ''
      @help_text = ''
      @supports_percent = false
    end

    def help_caption(text)
      @help_caption = text
    end

    def help_text(text)
      @help_text = text
    end

    def supports_percent
      @supports_percent = true
    end
  end

  cattr_accessor :operations
  self.operations = {}

  cattr_accessor :worker
  cattr_accessor :agentcfg
  cattr_accessor :config
  cattr_accessor :log

  attr_reader :plog

  attr_reader :uuid
  attr_reader :obj_type
  attr_reader :obj_id
  attr_reader :created_at
  attr_reader :expected_completion
  attr_reader :description
  attr_reader :subsystem
  attr_reader :agentnode
  attr_reader :orig_operation
  attr_reader :operation
  attr_reader :also_publish_to
  attr_reader :request_data

  attr_reader :is_tomte_rpc

  def self.operation(name, &block)
    op = Operation.new
    op.instance_exec(&block)

    operations[name.to_sym] = op
  end

  def initialize(request)
    request = OpenStruct.new(request) if !request.kind_of?(OpenStruct)

    @orig_operation = request.operation
    @operation = request.operation.downcase.to_sym

    if !self.class.operations[@operation]
      raise OperationNotDefined, "Operation #{request.operation} not defined"
    end

    @uuid = request.uuid
    @obj_type = request.obj_type
    @obj_id = request.obj_id
    @created_at = request.created_at
    @expected_completion = request.expected_completion
    @description = request.description
    @subsystem = request.subsystem
    @agentnode = request.agentnode
    @request_data = request.request_data

    @also_publish_to = request.also_publish_to

    @is_tomte_rpc = request.is_tomte_rpc

    @plog = BufferedLog.new(log)

    yield self if block_given?
  end

  def execute(opts = {}, &block)
    op = self.class.operations[operation]

    report_started(op.supports_percent ? { :percent => 0 } : {}) do
      @fiber = Fiber.new do
        begin
          if request_data.is_a?(Array)
            result_data = send(operation.downcase, *request_data)
          elsif method(operation.downcase).arity == 1
            result_data = send(operation.downcase, request_data)
          else
            result_data = send(operation.downcase)
          end
        rescue PermanentFailure => e
          plog.error e.message
          report_event(:PERMANENT_FAILURE, plog.text) { yield nil }
        rescue TemporaryFailure, PublishError => e
          plog.error e.message
          report_event(:TEMPORARY_FAILURE, plog.text) { yield nil }
        rescue Exception => e
          log.exception e
          plog.error e.message
          report_event(:TEMPORARY_FAILURE, plog.text) { yield nil }
        else
          report_event(:COMPLETED, plog.text, result_data) { yield result_data }
        end
      end

      @fiber.resume
    end
  end

  def human_output(data = '', &block)
    return unless current_request.payload.wants_human_output?
    current_request.response.payload.human_output ||= ''
    current_request.response.payload.human_output += data + (block ? block.call.to_s : '')
  end

  operation :banner do
    help_caption ''
    help_text ''
  end
  def banner
#    human_output do
#      <<-EOT.unindent
#      FIXME TODO
#
#      EOT
#    end

    { :api => 'FIXME', :class => self.class.name, :versions => { } }
  end

  operation :ping do
  end
  def ping
    :pong
  end

  operation :interface do
  end
  def interface
#    human_output do
#      'TODO'
#    end
    operations.inject_hash { |h, (v,m)| h[v] = m.opts.to_h }
  end



  def report_event(status, log, result_data = {}, &block)
    if also_publish_to
      also_publish_to.publish({
        :msg_type => status,
        :request_id => uuid,
        :result_data => result_data,
        :log => log,
      },
      :persistent => true,
      :errback => lambda { |a| raise PublishError },
      &block)
    else
      yield
    end
  end

  def report_started(opts = {}, &block)

    @lastprogress = Time.now

    if also_publish_to
      also_publish_to.publish({
        :msg_type => :STARTED,
        :request_id => uuid,
        :request_copy => {
          :obj_type => obj_type,
          :obj_id => obj_id,
          :created_at => created_at,
          :expected_completion => expected_completion,
          :description => description,
          :subsystem => subsystem,
          :agentnode => agentnode,
          :operation => orig_operation,
          :uuid => uuid,
          :request_data => request_data,
        },
      }.merge(opts),
      :persistent => true,
      :errback => lambda { |a| raise PublishError },
      &block)
    else
      yield
    end
  end

  def report_progress(percent)

    @lastprogress ||= Time.now

    if also_publish_to
      if Time.now - @lastprogress > 3
        also_publish_to.publish({
          :msg_type => :PROGRESS,
          :request_id => uuid,
          :percent => percent,
        },
        :errback => lambda { |a| raise PublishError },
        :persistent => true)

        @lastprogress = Time.now
      end
    end
  end
end

end
end
