#
# Copyright (C) 2013-2013, Intercom Srl, Daniele Orlandi
#
# Author:: Daniele Orlandi <daniele@orlandi.com>
#          Lele Forzani <lele@windmill.it>
#          Alfredo Cerutti <acerutti@intercom.it>
#
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.
#

require 'fiber'
require 'publisher'
require 'eventmachine'








module EventMachine
  def self.popen3(*args)
    new_stderr = $stderr.dup
    rd, wr = IO::pipe
    $stderr.reopen wr
    connection = EM.popen(*args)
    $stderr.reopen new_stderr
    EM.attach rd, Popen3StderrHandler, connection
    connection
  end
  
  class Popen3StderrHandler < EventMachine::Connection
    def initialize(connection)
      @connection = connection
    end
    
    def receive_data(data)
      @connection.receive_stderr(data)
    end
  end  
end


module Ygg
module Provisioner

module Tools
  def switch_files(f1, f2)
    begin ; File.unlink(f2 + '.old') rescue Errno::ENOENT ; end
    begin ; File.link(f2, f2 + '.old') rescue Errno::ENOENT ; end
    File.rename(f1, f2)
  end

  class CommandError < StandardError
    attr_accessor :status

    def initialize(msg, status)
      super msg
      self.status = status
    end
  end

  class CommandRunner < EventMachine::Connection
    extend Publisher

    can_fire :success, :failure, :receive_data, :receive_stderr

    # @private
    def initialize(collect_data = true)
      super
      @out = ''
      @collect_data = collect_data
    end

    def self.open(cmd, collect_data = true)
      EventMachine.popen3(cmd, self, collect_data)
    end

    def receive_data(data)
      @out += data if @collect_data
      fire :receive_data, data
    end

    def receive_stderr(data)
      fire :receive_stderr, data
    end

    def unbind
      if get_status.success?
        fire :success, @out, get_status
      else
        fire :failure, @out, get_status
      end
    end
  end

  def run_command(cmd, opts = {}, &block)
    if block
      run_command_async(cmd, opts, &block)

      nil
    else
      fiber = Fiber.current

      run_command(cmd, opts) do |out, status|
        fiber.resume(out, status)
      end

      (out, status) = Fiber.yield

      if opts[:with_status]
        [ out, status ]
      else
        raise CommandError.new('Error running command', status) if !status.success?
        out
      end
    end
  end

  #private
  def run_command_async(cmd, opts = {}, &block)
    out = ''

    plog = respond_to?(:plog) ? self.plog : opts[:plog]

    old_euid = Process::UID.eid
    old_egid = Process::GID.eid

    if opts[:uid]
      # This isn't the proper way to drop the privileges for the child process!!! FIXME TODO
      Process::UID.grant_privilege(opts[:uid])
    end

    if opts[:env]
      # This isn't the proper way to set the environment for the child process!!! FIXME TODO
      opts[:env].each { |k,v| ENV[k] = v }
    end

    plog.debug " = running command #{cmd.inspect}" if plog
    
    c = CommandRunner.open(cmd, !opts[:progress])

    if opts[:send_data]
      c.send_data(opts[:send_data])
    end

    if opts[:progress]
      c.on(:receive_data) do |out|
        yield out, nil if block_given?
      end

      c.on(:receive_stderr) do |out|
        opts[:stderr_handler].call(out) if opts[:stderr_handler]
      end
    end

    c.on(:success) do |out, result|
      out.each_line { |l| plog.info l } if plog

      yield out, result if block_given?
    end

    c.on(:failure) do |out, result|
      out.each_line { |l| plog.error l } if plog

      yield out, result if block_given?
    end

    Process::UID.eid = old_euid
    Process::GID.eid = old_egid

    nil
  end

end

end
end
